@@ -1,3 +1,4 @@
+/* $XFree86: xc/programs/Xserver/hw/xfree86/SuperProbe/OS_SYSV.c,v 3.15.2.2 1998/10/21 10:44:36 dawes Exp $ */
 /*
  * (c) Copyright 1993,1994 by David Wexelblat <dwex@xfree86.org>
  *
@@ -24,9 +25,7 @@
  * in this Software without prior written authorization from David Wexelblat.
  *
  */
-
-/* $XConsortium: OS_SYSV.c,v 1.5 95/01/16 13:16:18 kaleb Exp $ */
-/* $XFree86: xc/programs/Xserver/hw/xfree86/SuperProbe/OS_SYSV.c,v 3.5 1995/01/07 04:02:38 dawes Exp $ */
+/* $XConsortium: OS_SYSV.c /main/13 1996/10/25 11:33:20 kaleb $ */
 
 #include "Probe.h"
 
@@ -46,6 +45,9 @@
 #ifdef SCO
 # include <sys/vtkd.h>
 # include <sys/console.h>
+#ifdef SCO
+#undef _status
+#endif
 #else
 # include <sys/kd.h>
 # if !defined(sun)
@@ -56,10 +58,6 @@
 #if !defined(sun)
 # include <sys/region.h>
 #endif
-#if defined(sun)
-/* Fix for Solaris */
-#define __EXTENSIONS__
-#endif
 #include <sys/proc.h>
 #include <sys/tss.h>
 #ifdef NCR
@@ -83,24 +81,40 @@
 #ifdef __STDC__
 int sysi86(int, ...);
 int syscall(int, ...);
+#ifndef SCO
 int munmap(caddr_t, size_t);
+#endif
 #else
 int sysi86();
 int syscall();
+#ifndef SCO
 int munmap();
 #endif
+#endif
 
 #ifdef SVR4
 # include <sys/mman.h>
 # if defined(sun)
 #  define DEV_MEM	"/dev/fb"
 # else
-#  define DEV_MEM 	"/dev/pmem"
+#  ifdef SCO325
+#   define DEV_MEM	"/dev/mem"
+#  else
+#   define DEV_MEM 	"/dev/pmem"
+#  endif
 # endif
 #else
 # define DEV_MEM	"/dev/mem"
 #endif
 
+#ifndef SI86IOPL
+#define SET_IOPL() sysi86(SI86V86,V86SC_IOPL,PS_IOPL)
+#define RESET_IOPL() sysi86(SI86V86,V86SC_IOPL,0)
+#else
+#define SET_IOPL() sysi86(SI86IOPL,3)
+#define RESET_IOPL() sysi86(SI86IOPL,0)
+#endif
+
 static int VT_fd = -1;
 static int VT_num = -1;
 static int BIOS_fd = -1;
@@ -120,7 +134,7 @@ int OpenVideo()
 	if (geteuid() != 0)
 	{
 		fprintf(stderr,
-			"%s: Must be run as root or installed suid-root\n",
+			"%s: Must be run as root\n",
 			MyName);
 		return(-1);
 	}
@@ -167,7 +181,7 @@ int OpenVideo()
 		 */
 		sleep(1);
 	}
-	(void)sysi86(SI86V86, V86SC_IOPL, PS_IOPL);
+	(void)SET_IOPL();
 	return(VT_fd);
 }
 
@@ -181,7 +195,7 @@ void CloseVideo()
 {
 	if (VT_fd != -1)
 	{
-		(void)sysi86(SI86V86, V86SC_IOPL, 0);
+		(void)RESET_IOPL();
 	}
 	if (VT_fd > 0)
 	{
@@ -214,7 +228,7 @@ Byte *MapVGA()
 		fprintf(stderr, "%s: Failed to mmap framebuffer\n", MyName);
 		return((Byte *)0);
 	}
-#elif defined(SCO)
+#elif defined(OLDSCO)
 	static Byte *base = (Byte *)0;
 	ioctl(VT_fd, KDSETMODE, KD_GRAPHICS);
 	if (base == (Byte *) 0)
@@ -242,6 +256,13 @@ Byte *MapVGA()
 	return(base);
 }
 
+Byte *MapMem(address,size)
+	unsigned long address;
+	unsigned long size;
+{
+	return((Byte*)0);
+}
+
 /*
  * UnMapVGA --
  *
@@ -260,6 +281,13 @@ Byte *base;
 #endif
 }
 
+void UnMapMem(base,size)
+	Byte *base;
+	unsigned long size;
+{
+	return;
+}
+
 /*
  * ReadBIOS --
  *
