@@ -1,3 +1,4 @@
+/* $XConsortium: PatchLevel.h /main/15 1996/10/28 04:46:29 kaleb $ */
 /*
  * (c) Copyright 1993,1994 by David Wexelblat <dwex@xfree86.org>
  *
@@ -25,9 +26,10 @@
  *
  */
 
-/* $XConsortium: PatchLevel.h,v 1.4 95/01/12 19:16:57 kaleb Exp $ */
-/* $XFree86: xc/programs/Xserver/hw/xfree86/SuperProbe/PatchLevel.h,v 3.2 1994/11/30 20:36:13 dawes Exp $ */
+/* $XFree86: xc/programs/Xserver/hw/xfree86/SuperProbe/PatchLevel.h,v 3.20.2.5 1998/10/22 04:31:01 hohndel Exp $ */
 
 #define VERSION 	2
-#define PATCHLEV 	3
-#define PRINT_VERSION	printf("\n%s Version %d.%d\n",MyName,VERSION,PATCHLEV)
+#define PATCHLEV 	17
+#define RELDATE		"22 October 1998"
+#define PRINT_VERSION	printf("\n%s Version %d.%d (%s)\n", \
+				MyName,VERSION,PATCHLEV,RELDATE)
