@@ -1,6 +1,5 @@
 /*
- * $XConsortium: vgaHW.c,v 1.6 95/01/06 20:59:04 kaleb Exp $
- * $XFree86: xc/programs/Xserver/hw/xfree86/vga256/vga/vgaHW.c,v 3.15 1994/12/29 10:21:23 dawes Exp $
+ * $XFree86: xc/programs/Xserver/hw/xfree86/vga256/vga/vgaHW.c,v 3.50.2.5 1998/11/08 10:03:50 hohndel Exp $
  *
  * Copyright 1990,91 by Thomas Roell, Dinkelscherben, Germany.
  *
@@ -24,6 +23,7 @@
  *
  * Author:  Thomas Roell, roell@informatik.tu-muenchen.de
  */
+/* $XConsortium: vgaHW.c /main/19 1996/10/28 04:55:33 kaleb $ */
 
 #ifdef ISC202
 #include <sys/types.h>
@@ -38,7 +38,7 @@
 #include <sys/wait.h>
 #undef _POSIX_SOURCE
 #else
-#if defined(MINIX) || defined(AMOEBA)
+#if defined(MINIX) || defined(AMOEBA) || (defined(ISC) && defined(_POSIX_SOURCE)) || defined(Lynx)
 #include <sys/types.h>
 #endif
 #include <sys/wait.h>
@@ -56,19 +56,29 @@
 
 #include "xf86.h"
 #include "xf86Priv.h"
+#if 0
+#include <errno.h>
+#else
 #include "xf86_OSlib.h"
+#endif
 #include "xf86_HWlib.h"
 #include "vga.h"
+#ifdef PC98_EGC
+/* I/O port address define for extended EGC */
+#define		EGC_READ	0x4a2	/* EGC FGC,EGC,Read Plane  */
+#define		EGC_MASK	0x4a8	/* EGC Mask register       */
+#define		EGC_ADD		0x4ac	/* EGC Dest/Source address */
+#define		EGC_LENGTH	0x4ae	/* EGC Bit length          */
+#endif
 
-#ifndef MONOVGA
-#ifndef SCO
+#if !defined(PC98_NEC480) && !defined(PC98_EGC) && !defined(PC98_MGA)
+#if !defined(MONOVGA) && !defined(SCO)
 #ifndef SAVE_FONT1
 #define SAVE_FONT1
 #endif
 #endif
-#endif
 
-#if defined(CSRG_BASED) || defined(MACH386) || defined(linux) || defined(AMOEBA) || defined(MINIX)
+#if defined(Lynx) || defined(CSRG_BASED) || defined(MACH386) || defined(linux) || defined(AMOEBA) || defined(MINIX)
 #ifndef NEED_SAVED_CMAP
 #define NEED_SAVED_CMAP
 #endif
@@ -83,7 +93,7 @@
 #endif
 
 /* bytes per plane to save for text */
-#if defined(linux) || defined(MINIX)
+#if defined(Lynx) || defined(linux) || defined(MINIX)
 #define TEXT_AMOUNT 16384
 #else
 #define TEXT_AMOUNT 4096
@@ -91,6 +101,7 @@
 
 /* bytes per plane to save for font data */
 #define FONT_AMOUNT 8192
+#endif /* !defined(PC98_NEC480) && !defined(PC98_EGC) && !defined(PC98_MGA) */
 
 #if defined(CSRG_BASED) || defined(MACH386)
 #include <sys/time.h>
@@ -102,8 +113,6 @@
 #define WSTOPSIG(x) (x.w_stopsig)
 #endif
 
-extern void SetTimeSinceLastInputEvent();
-
 /* This the only where the definition seems to work (out of
  * vga.c/vgaHW.c/vgaCmap.c).
  */
@@ -116,17 +125,54 @@ Bool clgd6225Lcd= FALSE;
 #define OVERSCAN_VALUE 0x01
 #endif
 
+static CARD32
+vgaOffMode(
+#if NeedFunctionPrototypes
+     OsTimerPtr	/* timer */,
+     CARD32	/* now */,
+     pointer	/* arg */
+#endif
+);
+
+static CARD32
+vgaSuspendMode(
+#if NeedFunctionPrototypes
+     OsTimerPtr	/* timer */,
+     CARD32	/* now */,
+     pointer	/* arg */
+#endif
+);
+
+static Bool
+setExternClock(
+#if NeedFunctionPrototypes
+     int	/* clock */
+#endif
+);
+
 static int currentGraphicsClock = -1;
 static int currentExternClock = -1;
 
+Bool (*vgaBlankScreenFunc)()=vgaBlankScreen;
+
 int vgaRamdacMask = 0x3F;
-Bool vgaPowerSaver = FALSE;
 
 #define new ((vgaHWPtr)vgaNewVideoState)
 
 unsigned VGA_IOPorts[] = {
+#ifdef PC98_NEC480
+	0x60,  0x62,  0x6a,  0x7c,  0x7e,  0x80,  0xa0,  0xa2,  0x3c0, 0x3c2,
+	0x3c4, 0x3c6, 0x3c7, 0x3c8, 0x3c9, 0x3cc, 0x3ce, 0x3d0, 0x9a0, 0x9a8,
+#else
+#ifdef PC98_EGC
+/* I/O port address define Normal & Hireso mode */
+	0xa8,  0xaa,  0xac,  0xae,  0x4a0, 0x4a1, 0x4a2, 0x4a3, 0x4a4, 0x4a5,
+	0x4a6, 0x4a7, 0x4a8, 0x4a9, 0x4aa, 0x4ab, 0x4ac, 0x4ad, 0x4ae, 0x4af,
+#else
 	0x3B4, 0x3B5, 0x3BA, 0x3C0, 0x3C1, 0x3C2, 0x3C4, 0x3C5, 0x3C6, 0x3C7, 
 	0x3C8, 0x3C9, 0x3CA, 0x3CB, 0x3CC, 0x3CE, 0x3CF, 0x3D4, 0x3D5, 0x3DA,
+#endif
+#endif
 };
 int Num_VGA_IOPorts = (sizeof(VGA_IOPorts)/sizeof(VGA_IOPorts[0]));
 
@@ -203,33 +249,17 @@ unsigned char defaultDAC[768] =
 #endif /* NEED_SAVED_CMAP */
 
 /*
- * With USE_ASM_SLOWBCOPY, the version in common_hw/SlowBcopy.s is used.
+ * With Intel, the version in common_hw/SlowBcopy.s is used.
  * This avoids port I/O during the copy (which causes problems with
- * some hardware.
+ * some hardware).
  */
-#define USE_ASM_SLOWBCOPY
-#ifdef USE_ASM_SLOWBCOPY
-#define slowbcopy SlowBcopy
-#else
-/*
- * slowbcopy --
- *	slow version of bcopy for save/restore of font and text data.
- */
-static void
-slowbcopy(src, dst, count)
-     char *src, *dst;
-     int count;
-{
-  int i;
-  unsigned char temp;
-
-  for (i = 0; i < count; i++)
-  {
-    *dst++ = *src++;
-    outb(0x80, 0x00);
-  }
-}
-#endif
+#ifdef __alpha__
+#define slowbcopy_tobus(src,dst,count) SlowBCopyToBus(src,dst,count)
+#define slowbcopy_frombus(src,dst,count) SlowBCopyFromBus(src,dst,count)
+#else /* __alpha__ */
+#define slowbcopy_tobus(src,dst,count) SlowBcopy(src,dst,count)
+#define slowbcopy_frombus(src,dst,count) SlowBcopy(src,dst,count)
+#endif /* __alpha__ */
 
 /*
  * vgaProtect --
@@ -239,6 +269,7 @@ void
 vgaProtect(on)
      Bool on;
 {
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
   unsigned char tmp;
 
   if (xf86VTSema) {
@@ -249,7 +280,7 @@ vgaProtect(on)
       outb(0x3C4, 0x01);
       tmp = inb(0x3C5);
 
-      outw(0x3C4, 0x0100);			/* start synchronous reset */
+      (*vgaSaveScreenFunc)(SS_START);		/* start synchronous reset */
       outw(0x3C4, ((tmp | 0x20) << 8) | 0x01);	/* disable the display */
 
       tmp = inb(vgaIOBase + 0x0A);
@@ -263,65 +294,139 @@ vgaProtect(on)
       tmp = inb(0x3C5);
 
       outw(0x3C4, ((tmp & 0xDF) << 8) | 0x01);	/* reenable display */
-      outw(0x3C4, 0x0300);			/* clear synchronousreset */
+      (*vgaSaveScreenFunc)(SS_FINISH);		/* clear synchronousreset */
 
       tmp = inb(vgaIOBase + 0x0A);
       outb(0x3C0, 0x20);			/* disable pallete access */
     }
   }
-  return;
+#endif /* !defined(PC98_NEC480) && !defined(PC98_EGC) */
 }
 
 /*
- * vgaSaveScreen -- 
- *      Disable the video on the frame buffer to save the screen.
+ * vgaBlankScreen -- blank the screen.
  */
-Bool
-vgaSaveScreen (pScreen, on)
-     ScreenPtr     pScreen;
-     Bool          on;
+
+Bool vgaBlankScreen(ScreenPtr ptr, Bool on)
 {
-  unsigned char   state, state2;
 
-  if (on)
-    SetTimeSinceLastInputEvent();
-  if (xf86VTSema) {
-    outb(0x3C4,1);
-    state = inb(0x3C5);
-    outb(vgaIOBase + 4, 0x17);
-    state2 = inb(vgaIOBase + 5);
-  
-    if (on) {
-      state &= 0xDF;
-      state2 |= 0x80;
-    } else {
-      state |= 0x20;
-      state2 &= ~0x80;
-    }
-    
-    /*
-     * turn off screen if necessary
-     */
-    (*vgaSaveScreenFunc)(SS_START);
-    outw(0x3C4, 0x0100);              /* syncronous reset */
-    outw(0x3C4, (state << 8) | 0x01); /* change mode */
-    if (vgaPowerSaver) {
-      outb(vgaIOBase + 4, 0x17);
-      outb(vgaIOBase + 5, state2);
-    }
-    outw(0x3C4, 0x0300);              /* syncronous reset */
-    (*vgaSaveScreenFunc)(SS_FINISH);
+#if !defined(PC98_EGC) && !defined(PC98_PEGC)
+  unsigned char scrn;
 
-  } else {
-    if (on)
-      ((vgaHWPtr)vgaNewVideoState)->Sequencer[1] &= 0xDF;
-    else
-      ((vgaHWPtr)vgaNewVideoState)->Sequencer[1] |= 0x20;
+  outb(0x3C4,1);
+  scrn = inb(0x3C5);
+
+  if(on) {
+    scrn &= 0xDF;			/* enable screen */
+  }else {
+    scrn |= 0x20;			/* blank screen */
   }
 
-  return(TRUE);
+  (*vgaSaveScreenFunc)(SS_START);
+  outw(0x3C4, (scrn << 8) | 0x01); /* change mode */
+#if defined(Lynx) && defined(__i386__) && (defined(MONOVGA) || defined(XF86VGA16))
+  /* This helps to fix a complete system lockup using CT69000 VGA
+   * on LynxOS. I've seen this on two totally different systems.
+   * In each case there were no problems with VGA16 or VGA2 on
+   * FreeBSD while LynxOS alway crashed. Don't know what's going on
+   * with LynxOS....
+   */
+  usleep(1000);
+#endif
+  (*vgaSaveScreenFunc)(SS_FINISH);
+#else
+  if(on) 
+    outb(0xa2, 0xd);
+  else
+    outb(0xa2, 0xc);
+#endif    
+  return TRUE;
 }
 
+/*
+ * vgaSaveScreen -- blank the screen.
+ */
+
+Bool
+vgaSaveScreen(pScreen, on)
+     ScreenPtr pScreen;
+     Bool  on;
+{
+   if (on)
+      SetTimeSinceLastInputEvent();
+
+   if (xf86VTSema) {
+     (*vgaBlankScreenFunc)(pScreen,on);
+   }
+   return (TRUE);
+}
+
+/*
+ * vgaDPMSSet -- Sets VESA Display Power Management Signaling (DPMS) Mode
+ *
+ * This generic VGA function can only set the Off and On modes.  If the
+ * Standby and Suspend modes are to be supported, a chip specific replacement
+ * for this function must be written.
+ */
+
+void
+vgaDPMSSet(PowerManagementMode)
+    int PowerManagementMode;
+{
+#ifdef DPMSExtension
+  unsigned char seq1, crtc17;
+  if (!xf86VTSema) return;
+  switch (PowerManagementMode)
+  {
+  case DPMSModeOn:
+    /* Screen: On; HSync: On, VSync: On */
+    seq1 = 0x00;
+    crtc17 = 0x80;
+    break;
+  case DPMSModeStandby:
+    /* Screen: Off; HSync: Off, VSync: On -- Not Supported */
+    seq1 = 0x20;
+    crtc17 = 0x80;
+    break;
+  case DPMSModeSuspend:
+    /* Screen: Off; HSync: On, VSync: Off -- Not Supported */
+    seq1 = 0x20;
+    crtc17 = 0x80;
+    break;
+  case DPMSModeOff:
+    /* Screen: Off; HSync: Off, VSync: Off */
+    seq1 = 0x20;
+    crtc17 = 0x00;
+    break;
+  }
+  outw(0x3C4, 0x0100);	/* Synchronous Reset */
+  outb(0x3C4, 0x01);	/* Select SEQ1 */
+  seq1 |= inb(0x3C5) & ~0x20;
+  outb(0x3C5, seq1);
+  outb(vgaIOBase+4, 0x17); /* Select CRTC17 */
+  crtc17 |= inb(vgaIOBase+5) & ~0x80;
+  usleep(10000);
+  outb(vgaIOBase+5, crtc17);
+  outw(0x3C4, 0x0300);	/* End Reset */
+#endif
+}
+
+/*
+ * vgaHWSaveScreen
+ *      perform a sequencer reset.
+ */
+
+void
+vgaHWSaveScreen(start)
+    Bool start;
+{
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
+  if (start == SS_START)
+    outw(0x3C4, 0x0100);        /* synchronous reset */
+  else
+    outw(0x3C4, 0x0300);        /* end reset */
+#endif
+}
 
 /*
  * setExternClock
@@ -330,8 +435,8 @@ vgaSaveScreen (pScreen, on)
  */
 
 static Bool
-setExternClock(clock)
-     int clock;       /* the Clock index */
+setExternClock(clock2)
+     int clock2;       /* the Clock index */
 {
     int i;
 #ifdef MACH386
@@ -340,7 +445,7 @@ setExternClock(clock)
     int exit_status;
 #endif
 
-    if (clock == currentExternClock)
+    if (clock2 == currentExternClock)
       return(TRUE);
     switch(fork())
     {
@@ -369,14 +474,19 @@ setExternClock(clock)
         }
 #endif
         {
-          char *progname, clockarg[8], clockindex[3];
+          char *progname, clockarg[8], clockindex[4];
 
-          if (progname = rindex(vga256InfoRec.clockprog, '/'))
+          if ((progname = strrchr(vga256InfoRec.clockprog, '/')))
             progname++;
           else
             progname = vga256InfoRec.clockprog;
-          sprintf(clockarg, "%.3f", vga256InfoRec.clock[clock] / 1000.0);
-          sprintf(clockindex, "%d", clock);
+ 	  if(clock2 < MAXCLOCKS){
+ 	    sprintf(clockarg, "%.3f", vga256InfoRec.clock[clock2] / 1000.0);
+ 	    sprintf(clockindex, "%d", clock2 );
+ 	  } else {
+ 	    sprintf(clockarg, "%.3f", clock2 / 1000.0);
+ 	    sprintf(clockindex, "%d",255 );
+ 	  }
           execl(vga256InfoRec.clockprog, progname, clockarg, clockindex, NULL);
           ErrorF("Exec failed for ClockProg command \"%s\" (%s)\n",
                  vga256InfoRec.clockprog, strerror(errno));
@@ -420,7 +530,7 @@ setExternClock(clock)
           return(FALSE);
         }
     }
-    currentExternClock = clock;
+    currentExternClock = clock2;
     return(TRUE);
 }
 
@@ -436,8 +546,10 @@ vgaHWRestore(restore)
 {
   int i,tmp;
 
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
   tmp = inb(vgaIOBase + 0x0A);		/* Reset flip-flop */
   outb(0x3C0, 0x00);			/* Enables pallete access */
+#endif
 
   /*
    * This here is a workaround a bug in the kd-driver. We MUST explicitely
@@ -450,6 +562,7 @@ vgaHWRestore(restore)
    */
   
   vgaSaveScreen(NULL, FALSE);
+#if defined(SAVE_TEXT) || defined(SAVE_FONT1) || defined(SAVE_FONT2)
   if(restore->FontInfo1 || restore->FontInfo2 || restore->TextInfo) {
     /*
      * here we switch temporary to 16 color-plane-mode, to simply
@@ -464,53 +577,62 @@ vgaHWRestore(restore)
     outw(0x3CE,0xFF08); /* GJA - write all bits in a byte */
     outw(0x3CE,0x0001); /* GJA - all planes come from CPU */
 #endif
+#ifdef SAVE_FONT1
     if (restore->FontInfo1) {
       outw(0x3C4, 0x0402);    /* write to plane 2 */
       outw(0x3C4, 0x0604);    /* enable plane graphics */
       outw(0x3CE, 0x0204);    /* read plane 2 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy(restore->FontInfo1, vgaBase, FONT_AMOUNT);
+      slowbcopy_tobus(restore->FontInfo1, vgaBase, FONT_AMOUNT);
     }
+#endif
+#ifdef SAVE_FONT2
     if (restore->FontInfo2) {
       outw(0x3C4, 0x0802);    /* write to plane 3 */
       outw(0x3C4, 0x0604);    /* enable plane graphics */
       outw(0x3CE, 0x0304);    /* read plane 3 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy(restore->FontInfo2, vgaBase, FONT_AMOUNT);
+      slowbcopy_tobus(restore->FontInfo2, vgaBase, FONT_AMOUNT);
     }
+#endif
+#ifdef SAVE_TEXT
     if (restore->TextInfo) {
       outw(0x3C4, 0x0102);    /* write to plane 0 */
       outw(0x3C4, 0x0604);    /* enable plane graphics */
       outw(0x3CE, 0x0004);    /* read plane 0 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy(restore->TextInfo, vgaBase, TEXT_AMOUNT);
+      slowbcopy_tobus(restore->TextInfo, vgaBase, TEXT_AMOUNT);
       outw(0x3C4, 0x0202);    /* write to plane 1 */
       outw(0x3C4, 0x0604);    /* enable plane graphics */
       outw(0x3CE, 0x0104);    /* read plane 1 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy((char *)restore->TextInfo + TEXT_AMOUNT, vgaBase, TEXT_AMOUNT);
+      slowbcopy_tobus((char *)restore->TextInfo + TEXT_AMOUNT, vgaBase, TEXT_AMOUNT);
     }
+#endif
   }
+#endif /* defined(SAVE_TEXT) || defined(SAVE_FONT1) || defined(SAVE_FONT2) */
   vgaSaveScreen(NULL, TRUE);
 
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
+
   tmp = inb(vgaIOBase + 0x0A);			/* Reset flip-flop */
   outb(0x3C0, 0x00);				/* Enables pallete access */
 
-  if (vgaIOBase == 0x3B0)
-    restore->MiscOutReg &= 0xFE;
-  else
-    restore->MiscOutReg |= 0x01;
-
   /* Don't change the clock bits when using an external clock program */
   if (restore->NoClock < 0)
   {
     tmp = inb(0x3CC);
     restore->MiscOutReg = (restore->MiscOutReg & 0xF3) | (tmp & 0x0C);
   }
+  if (vgaIOBase == 0x3B0)
+    restore->MiscOutReg &= 0xFE;
+  else
+    restore->MiscOutReg |= 0x01;
+
   outb(0x3C2, restore->MiscOutReg);
 
   for (i=1; i<5;  i++) outw(0x3C4, (restore->Sequencer[i] << 8) | i);
@@ -548,6 +670,8 @@ vgaHWRestore(restore)
   tmp = inb(vgaIOBase + 0x0A);
   outb(0x3C0, 0x20);
 
+#endif /* !defined(PC98_NEC480) && !defined(PC98_EGC) */
+
   /* If restoring text mode, and a text clock is specified with clkprog */
   if (((restore->Attribute[0x10] & 0x01) == 0))
   {
@@ -581,10 +705,15 @@ vgaHWSave(save, size)
      int          size;
 {
   int           i,tmp;
-  Bool	        first_time = FALSE;
+  Bool	        first_time = FALSE;  /* Should be static? */
 
   if (save == NULL) {
-    save = (vgaHWPtr)Xcalloc(size);
+    tmp = size;
+    if (tmp < 0)
+      tmp = -size;
+    if (tmp < sizeof(vgaHWRec))
+      tmp = sizeof(vgaHWRec);
+    save = (vgaHWPtr)xcalloc(1,tmp);
     /*
      * Here we are, when we first save the videostate. This means we came here
      * to save the original Text mode. Because some drivers may depend
@@ -593,15 +722,18 @@ vgaHWSave(save, size)
     first_time = TRUE;
     if (vga256InfoRec.clockprog && vga256InfoRec.textclock >= 0)
       save->NoClock = -1;
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
     else
       /* This isn't very useful -- it ignores the high order CS bits */
       save->NoClock = (inb(0x3CC) >> 2) & 3;
+#endif /* !defined(PC98_NEC480) && !defined(PC98_EGC) */
   }
 
-  /*
-   * now get the fuck'in register
-   */
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
   save->MiscOutReg = inb(0x3CC);
+#ifdef PC98
+  save->MiscOutReg |= 0x01;
+#endif
   vgaIOBase = (save->MiscOutReg & 0x01) ? 0x3D0 : 0x3B0;
 
   tmp = inb(vgaIOBase + 0x0A); /* reset flip-flop */
@@ -635,13 +767,22 @@ vgaHWSave(save, size)
      */
     outb(0x3C7,0x00);
     for (i=0; i<3; i++) save->DAC[i] = inb(0x3C9);
+#if defined(PC98_PW) || defined(PC98_XKB) || defined(PC98_NEC) || defined(PC98_PWLB)
+    for (i=0; i<300; i++) inb(vgaIOBase + 4);
+#endif
     outb(0x3C8,0x00);
-    for (i=0; i<3; i++) outb(0x3C9, ~save->DAC[i]);
+    for (i=0; i<3; i++) outb(0x3C9, ~save->DAC[i] & vgaRamdacMask);
+#if defined(PC98_PW) || defined(PC98_XKB) || defined(PC98_NEC) || defined(PC98_PWLB)
+    for (i=0; i<300; i++) inb(vgaIOBase + 4);
+#endif
     outb(0x3C7,0x00);
     for (i=0; i<3; i++)
     {
-      unsigned char tmp = inb(0x3C9);
-      if (tmp != (~save->DAC[i] & vgaRamdacMask)) read_error++;
+      unsigned char tmp2 = inb(0x3C9);
+#if defined(PC98_PW) || defined(PC98_XKB) || defined(PC98_NEC) || defined(PC98_PWLB)
+      if (tmp2 != (~save->DAC[i]&0xFF))
+#endif
+      if (tmp2 != (~save->DAC[i] & vgaRamdacMask)) read_error++;
     }
   
     if (read_error)
@@ -681,16 +822,20 @@ vgaHWSave(save, size)
 
   for (i=0; i<5;  i++) { outb(0x3C4,i); save->Sequencer[i]   = inb(0x3C5); }
 
+#endif /* !defined(PC98_NEC480) && !defined(PC98_EGC) */
+
   vgaSaveScreen(NULL, FALSE);
   
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
   /* XXXX Still not sure if this is needed.  It isn't done in the Restore */
   outb(0x3C2, save->MiscOutReg | 0x01);		/* shift to colour emulation */
   /* Since forced to colour mode, must use 0x3Dx instead of (vgaIOBase + x) */
 
+#if defined(SAVE_TEXT) || defined(SAVE_FONT1) || defined(SAVE_FONT2)
   /*
    * get the character sets, and text screen if required
    */
-  if (((save->Attribute[0x10] & 0x01) == 0)) {
+  if (((save->Attribute[0x10] & 0x01) == 0) && (size >= 0)) {
 #ifdef SAVE_FONT1
     if (save->FontInfo1 == NULL) {
       save->FontInfo1 = (pointer)xalloc(FONT_AMOUNT);
@@ -707,7 +852,7 @@ vgaHWSave(save, size)
       outw(0x3CE, 0x0204);    /* read plane 2 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy(vgaBase, save->FontInfo1, FONT_AMOUNT);
+      slowbcopy_frombus(vgaBase, save->FontInfo1, FONT_AMOUNT);
     }
 #endif /* SAVE_FONT1 */
 #ifdef SAVE_FONT2
@@ -720,7 +865,7 @@ vgaHWSave(save, size)
       outw(0x3CE, 0x0304);    /* read plane 3 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy(vgaBase, save->FontInfo2, FONT_AMOUNT);
+      slowbcopy_frombus(vgaBase, save->FontInfo2, FONT_AMOUNT);
     }
 #endif /* SAVE_FONT2 */
 #ifdef SAVE_TEXT
@@ -737,24 +882,28 @@ vgaHWSave(save, size)
       outw(0x3CE, 0x0004);    /* read plane 0 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy(vgaBase, save->TextInfo, TEXT_AMOUNT);
+      slowbcopy_frombus(vgaBase, save->TextInfo, TEXT_AMOUNT);
       outw(0x3C4, 0x0202);    /* write to plane 1 */
       outw(0x3C4, 0x0604);    /* enable plane graphics */
       outw(0x3CE, 0x0104);    /* read plane 1 */
       outw(0x3CE, 0x0005);    /* write mode 0, read mode 0 */
       outw(0x3CE, 0x0506);    /* set graphics */
-      slowbcopy(vgaBase, (char *)save->TextInfo + TEXT_AMOUNT, TEXT_AMOUNT);
+      slowbcopy_frombus(vgaBase, (char *)save->TextInfo + TEXT_AMOUNT, TEXT_AMOUNT);
     }
 #endif /* SAVE_TEXT */
   }
+#endif /* defined(SAVE_TEXT) || defined(SAVE_FONT1) || defined(SAVE_FONT2) */
 
   outb(0x3C2, save->MiscOutReg);		/* back to original setting */
+#endif /* !defined(PC98_NEC480) && !defined(PC98_EGC) */
   
   vgaSaveScreen(NULL, TRUE);
 
+#if !defined(PC98_NEC480) && !defined(PC98_EGC)
   /* Turn on PAS bit */
   tmp = inb(vgaIOBase + 0x0A);
   outb(0x3C0, 0x20);
+#endif /* !defined(PC98_NEC480) && !defined(PC98_EGC) */
   
   return ((void *) save);
 }
@@ -772,7 +921,7 @@ vgaHWInit(mode, size)
      DisplayModePtr      mode;
      int             size;
 {
-  int                i;
+  unsigned int       i;
 
   /*
    * If we're using an external clock program, the first thing we do is
@@ -788,7 +937,9 @@ vgaHWInit(mode, size)
   }
 
   if (vgaNewVideoState == NULL) {
-    vgaNewVideoState = (void *)Xcalloc(size);
+    if (size < sizeof(vgaHWRec))
+      size = sizeof(vgaHWRec);
+    vgaNewVideoState = (void *)xcalloc(1,size);
 
     /*
      * initialize default colormap for monochrome
@@ -836,10 +987,17 @@ vgaHWInit(mode, size)
       }
       else
       {
-	if      (mode->CrtcVDisplay < 400) new->MiscOutReg = 0xA3;
-	else if (mode->CrtcVDisplay < 480) new->MiscOutReg = 0x63;
-	else if (mode->CrtcVDisplay < 768) new->MiscOutReg = 0xE3;
-	else                           new->MiscOutReg = 0x23;
+	int VDisplay = mode->VDisplay;
+	if (mode->Flags & V_DBLSCAN)
+	  VDisplay *= 2;
+	if      (VDisplay < 400)
+		new->MiscOutReg = 0xA3;		/* +hsync -vsync */
+	else if (VDisplay < 480)
+		new->MiscOutReg = 0x63;		/* -hsync +vsync */
+	else if (VDisplay < 768)
+		new->MiscOutReg = 0xE3;		/* -hsync -vsync */
+	else
+		new->MiscOutReg = 0x23;		/* +hsync +vsync */
       }
   if (!vga256InfoRec.clockprog)
     new->MiscOutReg |= (new->NoClock & 0x03) << 2;
@@ -852,7 +1010,10 @@ vgaHWInit(mode, size)
 #else
   new->Sequencer[0] = 0x00;
 #endif
-  new->Sequencer[1] = 0x01;
+  if (mode->Flags & V_CLKDIV2) 
+    new->Sequencer[1] = 0x09;
+  else
+    new->Sequencer[1] = 0x01;
 #ifdef MONOVGA
   new->Sequencer[2] = 1 << BIT_PLANE;
 #else
@@ -881,6 +1042,9 @@ vgaHWInit(mode, size)
   new->CRTC[1]  = (mode->CrtcHDisplay >> 3) - 1;
   new->CRTC[2]  = (mode->CrtcHSyncStart >> 3) -1;
   new->CRTC[3]  = ((mode->CrtcHSyncEnd >> 3) & 0x1F) | 0x80;
+  i = (((mode->CrtcHSkew << 2) + 0x10) & ~0x1F);
+  if (i < 0x80)
+    new->CRTC[3] |= i;
   new->CRTC[4]  = (mode->CrtcHSyncStart >> 3);
   new->CRTC[5]  = (((mode->CrtcHSyncEnd >> 3) & 0x20 ) << 2 )
     | (((mode->CrtcHSyncEnd >> 3)) & 0x1F);
@@ -909,7 +1073,7 @@ vgaHWInit(mode, size)
   new->CRTC[19] = vga256InfoRec.displayWidth >> 4;  /* just a guess */
   new->CRTC[20] = 0x00;
   new->CRTC[21] = mode->CrtcVSyncStart & 0xFF; 
-  new->CRTC[22] = (mode->CrtcVSyncStart +1) & 0xFF;
+  new->CRTC[22] = (mode->CrtcVSyncEnd + 1) & 0xFF;
 #if defined(MONOVGA) || defined(XF86VGA16)
   new->CRTC[23] = 0xE3;
 #else
@@ -981,6 +1145,32 @@ vgaHWInit(mode, size)
   new->Attribute[18] = 0x0F;
   new->Attribute[19] = 0x00;
   new->Attribute[20] = 0x00;
+#ifdef PC98_EGC
+  outb (0x7c, 0x00);
+  /* set to 16color mode */
+  outb(0x6a, 0x01);
+  /* set EGC enable */
+  outb(0x7c, 0xc0); /* GRCG enable, RMW mode */
+  outb(0x6a, 0x07);
+  outb(0x6a, 0x05);
+  outb(0x6a, 0x06);
+
+     for (i = 0; i < 4 ; i++)
+     {
+             outb (0x7e,0xff);
+     }
+
+       /* set up VGA registers */
+
+     outw (EGC_READ, 0x0000) ;
+     outw (EGC_MASK, 0xffff) ;
+     outw (EGC_READ ,0x40ff) ;
+     outw (EGC_ADD, 0x0000) ;
+     outw (EGC_LENGTH, 0x000f) ;
+#ifdef __linux__
+     outb(0x62, 0x0c);
+#endif /* __linux__ */
+#endif /* PC98_EGC */
   return(TRUE);
 }
 
@@ -991,10 +1181,13 @@ vgaHWInit(mode, size)
  */
 
 void
-vgaGetClocks(num, ClockFunc)
-     int num;
-     Bool (*ClockFunc)();
+vgaGetClocks(int num,
+#if NeedNestedPrototypes
+     Bool (*ClockFunc)(int))
+#else
+     Bool (*ClockFunc)())
+#endif
 {
-  xf86GetClocks(num, ClockFunc, vgaProtect, (void (*)())vgaSaveScreen,
+  xf86GetClocks(num, ClockFunc, vgaProtect, vgaSaveScreen,
 		(vgaIOBase + 0x0A), 0x08, 1, 28322, &vga256InfoRec);
 }
